import decimal
import json
import datetime

class ConversionError(ValueError): pass
class RequiredValueError(ValueError): pass

class Field:
    def __init__(self, name=None, ftype='object', required=False, **metadata):
        self.name = name
        self.ftype = ftype
        self.required = required
        self.null = None
        self.metadata = {'ftype': ftype, 'required': required, **metadata}

    def _is_required(self, value):
        if self.required and not value:
            raise RequiredValueError('Missing required value for {}'.format(self.name))
        else:
            return value

    def in_converter(self, value):
        if not self._is_required(value):
            return self.null
        else:
            return self._type_converter(value)

    def _type_converter(self, value):
        return value

    def __getitem__(self, key):
        return self.metadata[key]

    def __str__(self):
        return self.__dict__.__str__()

class StringField(Field):
    def __init__(self, name=None, length=None, **metadata):
        super().__init__(name, 'string', **metadata)
        self.length = length
        self.null = ''
        self.metadata = {**self.metadata, 'length': length, **metadata}

    def _type_converter(self, value):
        return str(value)

class IntegerField(Field):
    def __init__(self, name=None, **metadata):
        super().__init__(name, 'integer', **metadata)
        self.metadata = {**self.metadata, **metadata}

    def _type_converter(self, value):
        return int(value)

class DateField(Field):
    def __init__(self, name=None, in_format='%Y-%m-%d', out_format='%Y-%m-%d',**metadata):
        super().__init__(name, 'date', **metadata)
        self.in_format = in_format
        self.out_format = out_format
        self.metadata = {**self.metadata, 'in_format': in_format, 'out_format': out_format, **metadata}

    def _type_converter(self, value):
        if hasattr(value, 'date'):
            return value.date()
        elif hasattr(value, 'strftime'):
            return value
        else:
            return datetime.datetime.strptime(value, self.in_format).date()

class DateTimeField(Field):
    def __init__(self, name=None, in_format='%Y-%m-%d %H:%M:%S', out_format='%Y-%m-%d %H:%M:%S',**metadata):
        super().__init__(name, 'datetime', **metadata)
        self.in_format = in_format
        self.out_format = out_format
        self.metadata = {**self.metadata, 'in_format': in_format, 'out_format': out_format, **metadata}
        self.pd_converter = self.in_converter

    def _type_converter(self, value):
        return datetime.datetime.strptime(value, self.in_format)

class FloatField(Field):
    def __init__(self, name=None, **metadata):
        super().__init__(name, 'float', **metadata)
        self.metadata = {**self.metadata, **metadata}

    def _type_converter(self, value):
        return float(value)


# TODO: figure out a nice way for the converter to optionally truncate decimal without raising
class DecimalField(Field):
    def __init__(self, name=None, precision=None, scale=None, **metadata):
        super().__init__(name, 'decimal', **metadata)
        self.precision = precision
        self.scale = scale
        self.metadata = {**self.metadata, 'precision': precision, 'scale': scale, **metadata}

    def _type_converter(self, value):
        # TODO: This is just to catch an edge case where we get float('NaN') back from a db.  Refactor all None checking using pd.isnull()
        if value != value:
            return None

        dec = decimal.Decimal(str(value))

        detected_precision = len(dec.as_tuple().digits)
        detected_scale = -dec.as_tuple().exponent

        if detected_precision > self.precision:
            raise ConversionError('Decimal conversion error for "{}".  Expected precision: {}, Actual precision: {}'.format(
                value, self.precision, detected_precision
            ))
        if detected_scale > self.scale:
            raise ConversionError('Decimal conversion error for "{}".  Expected scale: {}, Actual scale: {}'.format(
                value, self.scale, detected_scale
            ))

        return dec

class BooleanField(Field):
    # TODOC: Point out that if unknown_truthiness is set, then that is used when no matching value is found (see tests)
    def __init__(self, name=None, map_true=[], map_false=[], **metadata):
        super().__init__(name, 'boolean', **metadata)
        self.metadata = {**self.metadata, **metadata}

        self.true_values = ['t','true','y','yes','on','1'] + map_true
        self.false_values = ['f','false','n','no','off','0'] + map_false

    def _type_converter(self, value):
        if value.lower() in self.true_values:
            return True
        elif value.lower() in self.false_values:
            return False
        elif 'unknown_truthiness' in self.metadata:
            return self.metadata['unknown_truthiness']
        elif not value:
            None
        else:
            raise ConversionError('Unable to convert "{}" to a boolean'.format(value))


class JsonField(Field):
    def __init__(self, name=None, **metadata):
        super().__init__(name, 'json', **metadata)
        self.metadata = {**self.metadata, **metadata}

    def _type_converter(self, value):
        return json.loads(value)

ftypes = {
   'object': Field,
   'string': StringField,
   'integer': IntegerField,
   'date': DateField,
   'datetime': DateTimeField,
   'float': FloatField,
   'decimal': DecimalField,
   'boolean': BooleanField,
   'json': JsonField
}



from collections import OrderedDict
import pandas as pd

class UncaughtMappingError(RuntimeError): pass
class PandasMapper:
    def __init__(self, df, *fieldmaps):
        self.df = df
        self.fieldmaps = fieldmaps

        self.mapped_df = pd.DataFrame([])
        self.mapping_errors = OrderedDict()
        self.errors_df = pd.DataFrame([])

    def execute(self, raise_errors=True):
        for fieldmap in self.fieldmaps:
            fieldmap.many_to_many_wrapper = self.many_to_many_wrapper()
            fieldmap.work_frame_builder = self.work_frame_builder

            fieldmap.apply(self.df, self.mapped_df)
            for idx, err in fieldmap.handler.errors.items():
                if idx not in self.mapping_errors:
                    self.mapping_errors[idx] = [[]]
                self.mapping_errors[idx][0].append(err)

        if raise_errors and len(self.mapping_errors) > 0:
            raise UncaughtMappingError(self.mapping_errors)
        else:
            return (self._mapped_df(), self._errors_df())

    def many_to_many_wrapper(self):
        return lambda row: pd.Series(row)

    def work_frame_builder(self, index):
        return pd.DataFrame(index=index)

    def _mapped_df(self):
        caught_idx = []
        for idx, errs in self.mapping_errors.items():
            has_caught = any([True for err in errs[0] if err['mode'] == 'catch'])
            if has_caught:
                caught_idx.append(idx)

        mapped_idx = [idx for idx in self.df.index if idx not in caught_idx]
        self.mapped_df = self.mapped_df.loc[mapped_idx].astype(object)
        return self.mapped_df


    def _errors_df(self):
        if len(self.mapping_errors) > 0:
            self.errors_df = self.df.loc[self.mapping_errors.keys()]
            self.errors_df['__mapping_errors__'] = pd.DataFrame(self.mapping_errors).transpose()[0].astype(object)

        return self.errors_df



from collections import OrderedDict

import pemi

class Handler:
    def __init__(self, mode='catch', recode=None):
        # modes: raise, catch, recode, warn

        self.mode = mode
        self.recode = recode
        self.errors = OrderedDict()

        self.apply_func = getattr(self, '_{}'.format(mode))

    def apply(self, func, row_arg, idx):
        try:
            return func(row_arg)
        except Exception as err:
            return self.apply_func(err, row_arg, idx)


    def mapping_error(self, err, idx):
        return {'mode': self.mode, 'index': idx, 'type': err.__class__.__name__, 'message': str(err)}

    def _raise(self, err, row_arg, idx):
        pemi.log().error(mapping_error(err, idx))
        raise err

    def _recode(self, err, row_arg, idx):
        self.errors[idx] = self.mapping_error(err, idx)
        return self.recode(row_arg)

    def _warn(self, err, row_arg, idx):
        pemi.log().warning(self.mapping_error(err, idx))
        return None

    def _catch(self, err, row_arg, idx):
        self.errors[idx] = self.mapping_error(err, idx)
        return None

class FieldMap:
    def __init__(self, source=None, target=None, transform=None, handler=Handler()):
        if isinstance(source, str):
            source = (source,)
        self.sources = source

        if isinstance(target, str):
            target = (target,)
        self.targets = target

        self.transform = transform
        self.handler = handler

        ## Must be set by the Mapper
        # Function to define how the many-to-many mapper converts a dict into a dataframe row
        self.many_to_many_wrapper = None
        # Function to define how to build a temp working dataframe (needs index)
        self.work_frame_builder = None

    def apply(self, source_df, target_df):
        work_df = self._build_work_df(source_df, target_df)

        if len(self.sources) == 1 and len(self.targets) == 1:
            target_df[self.targets[0]] = work_df.apply(self._one_to_one_map(), axis=1)
        elif len(self.targets) == 1:
            target_df[self.targets[0]] = work_df.apply(self._many_to_one_map(), axis=1)
        else:
            result_df = work_df.apply(self._many_to_many_map(), axis=1)
            for target in self.targets:
                target_df[target] = result_df[target]

        return None

    def _build_work_df(self, source_df, target_df):
        work_df = self.work_frame_builder(target_df.index)
        for source in self.sources:
            if source in target_df:
                work_df[source] = target_df[source]
            else:
                work_df[source] = source_df[source]
        return work_df

    def _one_to_one_map(self):
        func = lambda row: self.handler.apply(self.transform, row[self.sources[0]], row.name)
        return func

    def _many_to_one_map(self):
        func = lambda row: self.handler.apply(self.transform, row, row.name)
        return func

    def _many_to_many_map(self):
        wrapped = lambda inrow: self.many_to_many_wrapper(self.transform(inrow))
        func = lambda row: self.handler.apply(wrapped, row, row.name)
        return func



import unittest
from pandas.util.testing import assert_frame_equal

import pandas as pd

import pemi.mappers.pandas
from pemi.mappers.pandas import PandasMapper
from pemi import Handler
from pemi import FieldMap



def translate(val):
    if val == 1:
        return 'UNO'
    elif val == 2:
        return 'DOS'
    elif val == 3:
        return 'TRES'
    else:
        raise ValueError('Unknown translation: {}'.format(val))

def concatenate(delim=''):
    def doit(row):
        return delim.join(list(row.apply(str)))
    return doit

def mysplit(inrow):
    split_values = inrow['num_name'].split('-')
    outrow = { 'split_num': split_values[0], 'split_name': split_values[1] }
    return outrow

def unknown_recoder(val):
    return 'Unknown: {}'.format(val)


class TestCardinalities(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame(
            {
                'num': [1,2,3],
                'name': ['one', 'two', 'three'],
                'num_name': ['1-one', '2-two', '3-three']
            }
        )


    def test_one_to_one_map(self):
        '''
        One-to-one mapping
        '''
        mapper = PandasMapper(self.df,
            FieldMap(source='num', target='translated', transform=translate)
        )

        mapper.execute()
        result_df = mapper.mapped_df
        expected_df = pd.DataFrame({'translated': ['UNO', 'DOS', 'TRES']})

        assert_frame_equal(result_df, expected_df)

    def test_many_to_one_map(self):
        '''
        Many-to-one mapping
        '''

        mapper = PandasMapper(self.df,
            FieldMap(source=('name', 'num'), target='concatenated', transform=concatenate('-'))
        )

        mapper.execute()
        result_df = mapper.mapped_df
        expected_df = pd.DataFrame({'concatenated': ['one-1', 'two-2', 'three-3']})

        assert_frame_equal(result_df, expected_df)

    def test_one_to_many_map(self):
        '''
        One-to-many mapping
        '''

        mapper = PandasMapper(self.df,
            FieldMap(source='num_name', target=('split_name', 'split_num'), transform=mysplit)
        )

        mapper.execute()
        result_df = mapper.mapped_df
        expected_df = pd.DataFrame({
            'split_num': ['1', '2', '3'],
            'split_name': ['one', 'two', 'three']
        })

        assert_frame_equal(result_df, expected_df)

    def test_multiple_maps(self):
        '''
        Multiple mappings of various cardinalities
        '''
        mapper = PandasMapper(self.df,
            FieldMap(source='num', target='translated', transform=translate),
            FieldMap(source=('name', 'num'), target='concatenated', transform=concatenate('-')),
            FieldMap(source='num_name', target=('split_name', 'split_num'), transform=mysplit)
        )

        mapper.execute()
        result_df = mapper.mapped_df
        expected_df = pd.DataFrame(
            {
                'translated': ['UNO', 'DOS', 'TRES'],
                'concatenated': ['one-1', 'two-2', 'three-3'],
                'split_name': ['one', 'two', 'three'],
                'split_num': ['1', '2', '3']
            },
            columns=['translated', 'concatenated', 'split_name', 'split_num']
        )
        assert_frame_equal(result_df, expected_df)


class TestHandlerModes(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame(
            {
                'num': [1,20,3,40]
            }
        )

    def test_raise_mode(self):
        '''
        Raises an error immediately (default handler)
        '''
        mapper = PandasMapper(self.df,
            FieldMap(source='num', target='translated', transform=translate)
        )

        self.assertRaises(pemi.mappers.pandas.UncaughtMappingError, mapper.execute)


    def test_catch_mode(self):
        '''
        Errors are captured in an errors dataframe
        '''
        mapper = PandasMapper(self.df,
            FieldMap(source='num', target='translated', transform=translate, handler=Handler('catch'))
        )
        mapper.execute(raise_errors=False)
        mapped_df = mapper.mapped_df
        errors_df = mapper.errors_df

        expected_mapped_df = pd.DataFrame(
            {
                'translated': ['UNO','TRES']
            },
            index=[0,2]
        )

        expected_errors_df = pd.DataFrame(
            {
                'num': [20,40],
                '__mapping_errors__': [
                    [{'mode': 'catch', 'type': 'ValueError', 'message': 'Unknown translation: 20', 'index': 1}],
                    [{'mode': 'catch', 'type': 'ValueError', 'message': 'Unknown translation: 40', 'index': 3}]
                ]
            },
            index=[1,3],
            columns = ['num', '__mapping_errors__']
        )
        assert_frame_equal(mapped_df, expected_mapped_df)
        assert_frame_equal(errors_df, expected_errors_df)


    def test_catch_mode_raises(self):
        '''
        The catch handler will raise an error if errors are caught but not used
        '''
        mapper = PandasMapper(self.df,
            FieldMap(source='num', target='translated', transform=translate, handler=Handler('catch'))
        )
        self.assertRaises(pemi.mappers.pandas.UncaughtMappingError, mapper.execute)


    def test_recode_mode_one_to_one(self):
        '''
        The recode handler can be used to supply defaults for errors (one-to-one)
        '''
        def recoder(val):
            return 'Unknown: {}'.format(str(val))

        mapper = PandasMapper(self.df,
            FieldMap(source='num', target='translated', transform=translate, handler=Handler('recode', recode=recoder))
        )
        mapper.execute(raise_errors=False)
        mapped_df = mapper.mapped_df

        expected_mapped_df = pd.DataFrame(
            {
                'translated': ['UNO', 'Unknown: 20', 'TRES', 'Unknown: 40']
            },
        )
        assert_frame_equal(mapped_df, expected_mapped_df)


    def test_recode_mode_many_to_one(self):
        '''
        The recode handler can be used to supply defaults for errors (many-to-one)
        '''

        given_df = pd.DataFrame(
            {
                'num': ['1',2,'3'],
                'name': ['one', 'two', 'three']
            }
        )
        concatenate = lambda row: '='.join(list(row))
        recoder = lambda row: 'idk'

        mapper = PandasMapper(given_df,
            FieldMap(source=('num','name'), target=('combined'), transform=concatenate, handler=Handler('recode', recode=recoder))
        )
        mapper.execute(raise_errors=False)
        mapped_df = mapper.mapped_df

        expected_mapped_df = pd.DataFrame(
            {
                'combined': ['1=one', 'idk', '3=three']
            }
        )
        assert_frame_equal(mapped_df, expected_mapped_df)


    def test_recode_mode_many_to_many(self):
        '''
        The recode handler can be used to supply defaults for errors (many-to-many)
        '''
        given_df = pd.DataFrame(
            {
                'num_name': ['1-one', '2two', '3-three']
            }
        )

        def recoder(inrow):
            return {'split_num': None, 'split_name': None}

        mapper = PandasMapper(given_df,
            FieldMap(source='num_name', target=('split_num','split_name'), transform=mysplit, handler=Handler('recode', recode=recoder))
        )
        mapper.execute(raise_errors=False)
        mapped_df = mapper.mapped_df
        errors_df = mapper.errors_df

        expected_mapped_df = pd.DataFrame(
            {
                'split_num': ['1', None, '3'],
                'split_name': ['one', None, 'three']
            },
            columns = ['split_num', 'split_name']
        )
        assert_frame_equal(mapped_df, expected_mapped_df)

    def test_warn_mode(self):
        '''
        The warning handler effectively replaces errors with None
        '''
        mapper = PandasMapper(self.df,
            FieldMap(source='num', target='translated', transform=translate, handler=Handler('warn'))
        )
        mapper.execute(raise_errors=False)
        mapped_df = mapper.mapped_df
        errors_df = mapper.errors_df

        expected_mapped_df = pd.DataFrame(
            {
                'translated': ['UNO',None,'TRES',None]
            }
        )

        assert_frame_equal(mapped_df, expected_mapped_df)



import dask
import dask.dot

import pemi

class DaskPipe():
    'DaskPipe is just a wrapper around Pipe that allows us to use the Pipes in the context of Dask'

    def __init__(self, pipe):
        self.pipe = pipe

    def __call__(self, *args):
        pemi.log().info('DaskPipe flowing pipe {}'.format(self.pipe))
        self.pipe.flow()
        return self

    def __str__(self):
        return self.pipe.name

class DaskFlow():
    def __init__(self, connections):
        self.connections = connections

    def dag(self):
        dag = {}
        for conn in self.connections:
            for node, edge in self._node_edge(conn).items():
                if node in dag:
                    dag[node][1].extend(edge[1])
                else:
                    dag[node] = edge
        return dag

    def flow(self):
        all_nodes = list(self.dag().keys())
        non_recursive_nodes = [node for node in all_nodes if node != 'self.targets']
        return dask.get(self.dag(), non_recursive_nodes)

    def graph(self):
        return dask.dot.dot_graph(self.dag(), rankdir='TB')

    def _connect_to(self, source, connection):
        def __connect_to(target):
            pemi.log().debug('connecting {} to {}'.format(target, source))
            connection.connect()
            return target
        __connect_to.__name__ = 'connect_to'
        return __connect_to

    def _get_target(self, name):
        def __get_target(daskpipe):
            pemi.log().debug('Getting target {} from pipe {}'.format(name, daskpipe.pipe))
            return daskpipe.pipe.targets[name]
        __get_target.__name__ = '[{}]'.format(name)
        return __get_target

    def _node_edge(self, conn):
        return {
            '{}.targets'.format(conn.from_pipe.name): (DaskPipe(conn.from_pipe), []),
            '{}.targets[{}]'.format(conn.from_pipe.name, conn.from_subject.name): (self._get_target(conn.from_subject.name), '{}.targets'.format(conn.from_pipe.name)),
            '{}.sources[{}]'.format(conn.to_pipe.name, conn.to_subject.name): (self._connect_to(conn.to_pipe.sources[conn.to_subject.name], conn), '{}.targets[{}]'.format(conn.from_pipe.name, conn.from_subject.name)),
            '{}.targets'.format(conn.to_pipe.name): (DaskPipe(conn.to_pipe), ['{}.sources[{}]'.format(conn.to_pipe.name, conn.to_subject.name)])
        }



import pemi
from pemi import SourcePipe
from pemi import TargetPipe

from pemi.field_map import FieldMap
from pemi.field_map import Handler
from pemi.mappers.pandas import PandasMapper

import pandas as pd

class LocalCsvFileSourcePipe(SourcePipe):
    def __init__(self, **params):
        super().__init__(**params)

        self.schema = pemi.Schema(params['schema'])
        self.paths = params['paths']
        self.csv_opts = self._build_csv_opts(params.get('csv_opts', {}))

        self.targets['main'].schema = self.schema
        self.field_maps = self._build_field_maps()

    def extract(self):
        return self.paths

    def parse(self, filepaths):
        mapped_dfs = []
        error_dfs = []
        for filepath in filepaths:
            parsed_dfs = self._parse_one(filepath)
            mapped_dfs.append(parsed_dfs.mapped_df)
            error_dfs.append(parsed_dfs.errors_df)

        self.targets['main'].data = pd.concat(mapped_dfs)
        self.targets['error'].data = pd.concat(error_dfs)

        return self.targets['main'].data

    def _build_csv_opts(self, user_csv_opts):
        mandatory_opts = {
            'converters': self.schema.str_converters(),
            'usecols':    self.schema.keys()
        }

        default_opts = {
            'engine':          'c',
            'error_bad_lines': True
        }

        return {**default_opts, **user_csv_opts, **mandatory_opts}

    def _build_field_maps(self):
        field_maps = []
        for name, field in self.schema.items():
            fm = FieldMap(source=name, target=name, transform=field.in_converter, handler=Handler('catch'))
            field_maps.append(fm)

        return field_maps

    def _parse_one(self, filepath):
        raw_df = pd.read_csv(filepath, **self.csv_opts)
        mapper = PandasMapper(raw_df, *self.field_maps)
        mapper.execute(raise_errors=False)
        return mapper


class LocalCsvFileTargetPipe(TargetPipe):
    def __init__(self, **params):
        super().__init__(**params)

        self.schema = pemi.Schema(params['schema'])
        self.path = params['path']
        self.csv_opts = self._build_csv_opts(params.get('csv_opts', {}))

    def encode(self):
        return self.sources['main'].data

    def load(self, df):
        df.to_csv(self.path, **self.csv_opts)
        return self.path

    def _build_csv_opts(self, user_csv_opts):
        mandatory_opts = {}

        default_opts = {
            'index': False
        }

        return {**default_opts, **user_csv_opts, **mandatory_opts}



from collections import OrderedDict

import pemi
from pemi.data_subject import PdDataSubject

class PipeConnection():
    def __init__(self, parent, from_subject):
        self.parent = parent
        self.from_pipe_name = from_subject.pipe.name
        self.from_subject_name = from_subject.name

    def to(self, to_subject):
        self.to_pipe_name = to_subject.pipe.name
        self.to_subject_name = to_subject.name
        return self

    @property
    def from_pipe(self):
        return self.parent.pipes[self.from_pipe_name]

    @property
    def to_pipe(self):
        return self.parent.pipes[self.to_pipe_name]

    @property
    def from_subject(self):
        return self.from_pipe.targets[self.from_subject_name]

    @property
    def to_subject(self):
        return self.to_pipe.sources[self.to_subject_name]

    def connect(self):
        self.to_subject.connect_from(self.from_subject)

    def __str__(self):
        return 'PipeConnection: {}.{} -> {}.{}'.format(
            self.from_pipe,
            self.from_subject,
            self.to_pipe,
            self.to_subject
        )

    def __repr__(self):
        return '<{}>'.format(self.__str__())


class Pipe():
    '''
    A pipe is a parameterized collection of sources and targets which can be executed (flow).
    '''

    def __init__(self, name=None, **params):
        self.name = name
        self.params = params
        self.sources = OrderedDict()
        self.targets = OrderedDict()
        self.pipes = OrderedDict()
        self.connections = []
        self.config()

    def config(self):
        'Override this to configure attributes of specific pipes (sources, targets, connections, etc)'
        pass

    def source(self, subject_class, name, schema=pemi.Schema(), **kwargs):
        self.sources[name] = subject_class(
            pipe=self,
            name=name,
            schema=schema,
            **kwargs
        )

    def target(self, subject_class, name, schema=pemi.Schema(), **kwargs):
        self.targets[name] = subject_class(
            pipe=self,
            name=name,
            schema=schema,
            **kwargs
        )

    def pipe(self, name, pipe):
        pipe.name = name
        self.pipes[name] = pipe


    def connect(self, connect_from):
        conn = PipeConnection(self, connect_from)

        self.connections.append(conn)
        return conn


    def flow(self):
        raise NotImplementedError

    def __str__(self):
        return "<{}({}) {}>".format(self.__class__.__name__, self.name, id(self))


class SourcePipe(Pipe):
    '''
    A source pipe extracts data from some external source, and parses
    it into a data structure that can be used by subsequent pipes.
    '''

    def __init__(self, subject_class=PdDataSubject, **params):
        super().__init__(**params)

        self.target(subject_class, name='main')
        self.target(subject_class, name='error')

    def extract(self):
        #e.g., S3SourceExtractor.extract()
        raise NotImplementedError
        return extracted_data_that_can_be_parsed

    def parse(self, data):
        #e.g., CsvParser.parse(data)
        raise NotImplementedError
        return parsed_data

    def flow(self):
        self.targets['main'].data = self.parse(self.extract())
        return self.targets['main'].data


class TargetPipe(Pipe):
    '''
    A target pipe takes data provided to it, encodes it into a structure that can be
    understand by some external target, and then loads the data into that external target.
    '''

    def __init__(self, subject_class=PdDataSubject, **params):
        super().__init__(**params)

        self.source(subject_class, name='main')
        self.target(subject_class, name='load_response')

    def encode(self):
        #e.g., CsvTargetEncoder.encode()
        raise NotImplementedError
        return source_main_data_encoded_for_loader

    def load(self, encoded_data):
        #e.g., S3Loader.load()
        raise NotImplementedError
        return results_from_load_operation

    def flow(self):
        return self.load(self.encode())



'''
Package containing core Pemi functionality
'''

import logging
logging.basicConfig()
logging.getLogger('pemi').setLevel(logging.WARN)
def log(name='pemi'):
    return logging.getLogger(name)


from pemi.schema import Schema
from pemi.field import Field

from pemi.pipes.pipe import Pipe
from pemi.pipes.pipe import SourcePipe
from pemi.pipes.pipe import TargetPipe

from pemi.field_map import FieldMap
from pemi.field_map import Handler



import pemi.field

from collections import OrderedDict

class Schema:
    def __init__(self, schema={}):
        if schema.__class__ == Schema:
            self.fields = schema.fields
        else:
            self.fields = self.__fields_from_dict(schema)


    def __fields_from_dict(self, schema_dict):
        fields_dict = OrderedDict()

        for name, meta in schema_dict.items():
            init_meta = {k:v for (k,v) in meta.items() if k != 'ftype'}
            fields_dict[name] = pemi.field.ftypes[meta['ftype']](name, **init_meta)
        return fields_dict

    def __getitem__(self, key):
        return self.fields[key]

    def keys(self):
        return self.fields.keys()

    def values(self):
        return self.fields.values()

    def items(self):
        return self.fields.items()

    def in_converters(self):
        return {f.name: f.in_converter for f in self.fields.values()}

    def str_converters(self):
        return {f.name: str for f in self.fields.values()}

    def __str__(self):
        return "\n".join(['{} -> {}'.format(name, meta.__str__()) for name, meta in self.fields.items()])

    def merge(self, other):
        merged_fields = {**self.fields, **other.fields}
        for name, field in merged_fields.items():
            if name in self.fields:
                self_meta = self.fields[name].metadata
            else:
                self_meta = {}

            if name in other.fields:
                other_meta = other.fields[name].metadata
            else:
                other_meta = {}

            merged_field_metadata = {**self_meta, **other_meta}
            merged_fields[name] = merged_field_metadata

        return Schema(merged_fields)
